import functools
import inspect

from . import _decoding_
from . import _schema_

__all__ = ["validate_execution"]


def validate_execution(
    function,
    *,
    validate_payload: bool = True,
    validate_return: bool = True,
):
    payload_annotation, return_annotation = _schema_.extract_annotations(function)

    if not validate_payload or payload_annotation is ...:
        payload_validator = lambda v: v
    else:
        payload_validator = _decoding_.serialize(payload_annotation)

    if not validate_return or return_annotation is ...:
        # raise RuntimeError(f'set {procedure} returning annotation')
        return_validator = lambda v: v
    else:
        return_validator = _decoding_.serialize(return_annotation)

    @functools.wraps(function)
    async def async_decorator(payload, *args, **kwargs):
        payload = payload_validator(payload)
        result = await function(payload, *args, **kwargs)
        return return_validator(result)

    if inspect.iscoroutinefunction(function):
        return async_decorator

    @functools.wraps(function)
    def decorator(payload, *args, **kwargs):
        payload = payload_validator(payload)
        result = function(payload, *args, **kwargs)
        return return_validator(result)

    return decorator



import asyncio
import logging
import re
import typing

from . import _service_
from . import _shared_

if typing.TYPE_CHECKING:
    from . import _almanet_

__all__ = [
    "observable_state",
    "transition",
    "next_observer",
]


_logger = logging.getLogger(__name__)


@_shared_.dataclass(slots=True)
class transition:
    label: str
    source: "observable_state"
    target: "observable_state"
    procedure: typing.Callable
    description: str | None = None
    priority: int = -1
    is_async: bool = False

    def __post_init__(self):
        if asyncio.iscoroutinefunction(self.procedure):
            self.is_async = True

    def __call__(
        self,
        session: "_almanet_.Almanet",
        *args,
        **kwargs,
    ):
        if self.is_async:
            coroutine = self.procedure(*args, **kwargs, transition=self)
            task = asyncio.create_task(coroutine)
            task.add_done_callback(
                lambda task: self.target.notify(session, task.result())
            )
            return task

        result = self.procedure(*args, **kwargs, transition=self)
        self.target.notify(session, result)
        return result

    def __str__(self) -> str:
        return f"{self.label}: {self.source.label} > {self.target.label}"

    def __hash__(self) -> int:
        return hash(self.__str__())


_state_label_re = re.compile("[A-Za-z_]+")


@_shared_.dataclass(slots=True)
class _state:

    entity: str
    label: str
    service: _service_.service
    description: str | None = None
    transitions: typing.List[transition] = ...

    def __post_init__(self):
        if not (isinstance(self.label, str) and len(self.label) > 0 and _state_label_re.match(self.label)):
            raise ValueError("`label` must contain uppercase words separated by underscore")
        self.transitions = []

    def _add_transition(
        self,
        label: str | None,
        description: str | None,
        target: "observable_state",
        procedure: typing.Callable,
        **extra,
    ) -> transition:
        if not callable(procedure):
            raise ValueError("decorated function must be callable")

        if label is None:
            label = f"{procedure.__module__}.{procedure.__name__}"

        if description is None:
            description = procedure.__doc__

        if not isinstance(target, observable_state):
            raise ValueError(f"{label}: `target` must be `observable_state` instance")

        instance = transition(
            label=label,
            description=description,
            source=self, # type: ignore
            target=target,
            procedure=procedure,
            **extra,
        )
        self.transitions.append(instance)
        return instance

    def transition(
        self,
        target: "observable_state",
        label: str | None = None,
        description: str | None = None,
        **extra,
    ):
        def wrap(function):
            return self._add_transition(
                label=label,
                description=description,
                target=target,
                procedure=function,
                **extra,
            )
        return wrap


LOWEST_PRIORITY = 100
LOW_PRIORITY = 75
MEDIUM_PRIORITY = 50
HIGH_PRIORITY = 25
HIGHEST_PRIORITY = 0

DEFAULT_PRIORITY = MEDIUM_PRIORITY


@_shared_.dataclass(slots=True)
class observable_state(_state):

    _observers: typing.List[transition] = ...

    @property
    def route(self) -> str:
        return f'{self.entity}.{self.label}.next'

    @property
    def full_route(self) -> str:
        if isinstance(self.service.pre, str):
            return f'{self.service.pre}.{self.route}'
        return self.route

    async def next(
        self,
        payload: typing.Any,
        session: "_almanet_.Almanet",
        **kwargs,
    ) -> typing.Any:
        _logger.debug(f"{self.label} begin")

        if payload is None:
            payload = []
        if isinstance(payload, typing.Iterable):
            payload = [payload]

        for observer in self._observers:
            _logger.debug(f"trying to call {observer.label} observer")
            try:
                if observer.is_async:
                    result = await observer(session, *payload, **kwargs)
                else:
                    result = await asyncio.to_thread(observer, session, *payload, **kwargs)
                _logger.debug(f"{observer.label} observer end")
                return result
            except Exception as e:
                _logger.error(f"during execution of {observer.label}: {repr(e)}")

    def __post_init__(self):
        super(observable_state, self).__post_init__()
        self._observers = []
        self.service.add_procedure(self.next, label=self.route, include_to_api=False, validate=False)

    def notify(
        self,
        session: "_almanet_.Almanet",
        previous_result: typing.Any,
    ):
        return session.call(
            self.full_route,
            previous_result,
        )

    def _add_observer(
        self,
        priority: int,
        **kwargs,
    ) -> transition:
        if not (isinstance(priority, int) and priority > -1):
            raise ValueError("`priority` must be `integer` and greater than -1")

        instance = self._add_transition(**kwargs, priority=priority)

        existing_priorities = {i.priority for i in self._observers}
        if priority in existing_priorities:
            raise ValueError(f"invalid observer `{instance.label}`: priority {priority} already exists")

        self._observers.append(instance)
        self._observers = sorted(self._observers, key=lambda x: -x.priority)

        return instance

    def observer(
        self,
        target: "observable_state",
        label: str | None = None,
        description: str | None = None,
        priority: int = DEFAULT_PRIORITY,
        **extra,
    ):
        def wrap(function):
            return self._add_observer(
                label=label,
                description=description,
                target=target,
                procedure=function,
                priority=priority,
                **extra,
            )
        return wrap


class next_observer(Exception):

    def __init__(self, reason: str, *args):
        if not isinstance(reason, str):
            raise ValueError("`reason` must be `string`")
        self.reason = reason
        self.args = args



import asyncio

__all__ = [
    "merge_streams",
    "make_closable"
]


async def merge_streams(
    *streams
):
    pending_tasks = [asyncio.create_task(anext(i)) for i in streams]
    active = True
    while active:
        done_tasks, _ = await asyncio.wait(pending_tasks, return_when=asyncio.FIRST_COMPLETED)
        for dt in done_tasks:
            result = dt.result()
            if isinstance(result, StopAsyncIteration):
                active = False
                continue
            yield result

            i = -1
            for pt in pending_tasks:
                i += 1
                if pt is not dt:
                    continue

                s = streams[i]
                pending_tasks[i] = asyncio.create_task(anext(s))
                break


def make_closable(
    stream,
    on_close = None,
):
    run_callback = asyncio.iscoroutinefunction(on_close)

    close_event = asyncio.Event()

    async def on_close_stream():
        await close_event.wait()
        if run_callback:
            await on_close()
        yield StopAsyncIteration()

    new_stream = merge_streams(stream, on_close_stream())
    return new_stream, close_event.set



from dataclasses import dataclass

from ._decoding_ import *
from ._encoding_ import *
from ._new_id_ import *
from ._schema_ import *
from ._streaming_ import *
from ._task_pool_ import *
from ._validation_ import *
from ._uri_ import *

__all__ = [
    "dataclass",
    *_decoding_.__all__,
    *_encoding_.__all__,
    *_new_id_.__all__,
    *_schema_.__all__,
    *_streaming_.__all__,
    *_task_pool_.__all__,
    *_validation_.__all__,
    *_uri_.__all__,
]



import typing
from ._almanet_ import *
from . import _clients_ as clients
from ._flow_ import *
from ._service_ import *

__all__ = [
    *_almanet_.__all__,
    "clients",
    'new_session',
    *_flow_.__all__,
    *_service_.__all__,
    'new_service',
]


def new_session(
    *addresses: str,
    client_klass: type[client_iface] = clients.DEFAULT_CLIENT,
    **kwargs: typing.Unpack[Almanet._kwargs],
) -> Almanet:
    return Almanet(
        *addresses,
        client=client_klass(),
        **kwargs
    )


def new_service(
    *addresses: str,
    session: Almanet | None = None,
    **kwargs: typing.Unpack[service._kwargs],
) -> service:
    if session is None:
        session = new_session(*addresses)
    return service(session, **kwargs)



import asyncio
import typing

from . import _almanet_
from . import _shared_

__all__ = [
    "service",
]


class service:

    class _kwargs(typing.TypedDict):
        prefix: typing.NotRequired[str]
        tags: typing.NotRequired[typing.Set[str]]

    def __init__(
        self,
        session: _almanet_.Almanet,
        **kwargs: typing.Unpack[_kwargs],
    ):
        self._post_join_callbacks = []
        self._routes = set()
        self.pre = kwargs.get('prefix')
        self.tags = set(kwargs.get('tags') or [])
        self.session = session

    async def _share_self_schema(
        self,
        **extra,
    ):
        async def procedure(*args, **kwargs):
            return {
                'client': self.session.id,
                'version': self.session.version,
                'routes': list(self._routes),
                **extra,
            }

        await self.session.register(
            '_api_schema_.client',
            procedure,
            channel=self.session.id,
        )

    async def _share_procedure_schema(
        self,
        topic: str,
        channel: str,
        tags: set[str] | None = None,
        **extra,
    ) -> None:
        if tags is None:
            tags = set()
        tags |= self.tags
        if len(tags) == 0:
            tags = {'Default'}

        async def procedure(*args, **kwargs):
            return {
                'client': self.session.id,
                'version': self.session.version,
                'topic': topic,
                'channel': channel,
                'tags': tags,
                **extra,
            }

        await self.session.register(
            f'_api_schema_.{topic}.{channel}',
            procedure,
            channel=channel,
        )

        self._routes.add(f'{topic}/{channel}')

    class _register_procedure_kwargs(typing.TypedDict):
        label: typing.NotRequired[str]
        channel: typing.NotRequired[str]
        validate: typing.NotRequired[bool]
        include_to_api: typing.NotRequired[bool]
        title: typing.NotRequired[str]
        description: typing.NotRequired[str]
        tags: typing.NotRequired[set[str]]
        payload_model: typing.NotRequired[typing.Any]
        return_model: typing.NotRequired[typing.Any]

    async def _register_procedure(
        self,
        procedure: typing.Callable,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ):
        topic = kwargs.get('label', procedure.__name__)

        if isinstance(self.pre, str):
            topic = f'{self.pre}.{topic}'

        if kwargs.get('validate', True):
            procedure = _shared_.validate_execution(procedure)

        registration = await self.session.register(topic, procedure, channel=kwargs.get('channel'))

        if kwargs.get('include_to_api', True):
            procedure_schema = _shared_.describe_function(
                procedure,
                kwargs.get('description'),
                kwargs.get('payload_model', ...),
                kwargs.get('return_model', ...),
            )
            await self._share_procedure_schema(
                topic,
                registration.channel,
                title=kwargs.get('title'),
                tags=kwargs.get('tags'),
                **procedure_schema,
            )

    def add_procedure(
        self,
        procedure: typing.Callable,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> None:
        self._post_join_callbacks.append(
            lambda: self._register_procedure(procedure, **kwargs)
        )

    def procedure(
        self,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ):
        def decorate(function):
            self.add_procedure(function, **kwargs)
            return function
        return decorate

    async def _post_serve(self):
        await self.session.join()

        for callback in self._post_join_callbacks:
            coroutine = callback()
            self.session.task_pool.schedule(coroutine)

        await self._share_self_schema()

    def serve(self):
        loop = asyncio.new_event_loop()
        loop.create_task(
            self._post_serve()
        )
        loop.run_forever()



import typing

import pydantic


__all__ = [
    "extract_annotations",
    "generate_json_schema",
    "describe_function",
]


def extract_annotations(
    function: typing.Callable,
):
    payload_annotation = function.__annotations__.get('payload', ...)
    return_annotation = function.__annotations__.get('return', ...)
    return payload_annotation, return_annotation


def generate_json_schema(annotation):
    """
    Generates a JSON schema from an annotation.
    """
    if annotation is ...:
        return None

    model = pydantic.TypeAdapter(annotation)
    return model.json_schema()


def describe_function(
    f: typing.Callable,
    description: str | None = None,
    payload_model: typing.Any = ...,
    return_model: typing.Any = ...,
):
    """
    Returns a dictionary containing the description, payload model, and return model of the function.
    """
    if description is None:
        description = f.__doc__

    if payload_model is ...:
        payload_annotation = f.__annotations__.get('payload', ...)
    else:
        payload_annotation = payload_model

    payload_json_schema = generate_json_schema(payload_annotation)

    if return_model is ...:
        return_annotation = f.__annotations__.get('return', ...)
    else:
        return_annotation = return_model

    return_json_schema = generate_json_schema(return_annotation)

    return {
        'description': description,
        'payload_json_schema': payload_json_schema,
        'return_json_schema': return_json_schema,
    }
